"""Top-level package for metamock."""

from .cli import cli
from .version import __author__, __email__, __timestamp__, __version__

__all__ = ["cli", __version__, __timestamp__, __author__, __email__]



import os.path
import subprocess
from datetime import datetime

import boto
import boto.iam
import boto.sts

from . import otp
from .util import cache, first_item, get_value


class NoSuchProfileError(Exception):
    def __init__(self, name):
        super(NoSuchProfileError, self).__init__(
            "Profile {} not found.".format(name)
        )


class Metadata(object):
    def __init__(self, profiles, profile_name):
        self.profiles = profiles
        self.profile_name = profile_name

    @property
    def profile_name(self):
        return self.__profile_name

    @profile_name.setter
    def profile_name(self, name):
        if name not in self.profiles:
            raise NoSuchProfileError(name)

        self.__profile_name = name

    @property
    def profile(self):
        return self.profiles[self.profile_name]

    # Replace uses of the following methods with metadata.profile.x

    @property
    def session(self):
        return self.profile.session

    @property
    def session_expired(self):
        return self.profile.session_expired

    def clear_session(self, *args, **kw):
        return self.profile.clear_session(*args, **kw)

    def get_session(self, *args, **kw):
        return self.profile.get_session(*args, **kw)


class Profile(object):
    def __init__(
        self,
        region="us-east-1",
        access_key=None,
        secret_key=None,
        token_duration=None,
        role_arn=None,
        mfa_secret=None,
        mfa_enabled=True,
    ):
        self.region = region
        self.access_key = access_key
        self.secret_key = secret_key
        self.token_duration = token_duration
        self.role_arn = role_arn
        self.mfa_enabled = mfa_enabled
        self.otp = otp.Totp(mfa_secret) if mfa_secret is not None else None

        self.user = None
        self.session = None
        self.session_expiration = datetime.min
        self.mfa_serial_number = None

        self._load_user()

        if role_arn:
            self.role_arn = role_arn.replace(
                "${aws:username}", self.user.user_name
            )

    @cache
    def iam(self):
        return boto.iam.connect_to_region(
            self.region,
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
        )

    @cache
    def sts(self):
        return boto.sts.connect_to_region(
            self.region,
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
        )

    @property
    def session_expired(self):
        return datetime.utcnow() > self.session_expiration

    def _load_user(self):
        self.user = get_value(
            self.iam.get_user(), "get_user_response", "get_user_result", "user"
        )

        if self.mfa_enabled:
            self._load_mfa_device()

    def _load_mfa_device(self):
        mfa_device = first_item(
            get_value(
                self.iam.get_all_mfa_devices(self.user.user_name),
                "list_mfa_devices_response",
                "list_mfa_devices_result",
                "mfa_devices",
            )
        )

        self.mfa_serial_number = (
            mfa_device.serial_number if mfa_device else None
        )

    def _prompt_token(self):
        if self.otp:
            return str(self.otp.generate())
        else:
            script = os.path.join(os.path.dirname(__file__), "prompt.py")
            return subprocess.check_output(["/usr/bin/python", script]).strip()

    def _create_session(self, token_value):
        if self.role_arn:
            role = self.sts.assume_role(
                role_arn=self.role_arn,
                role_session_name="some_session_id",
                duration_seconds=self.token_duration,
                mfa_serial_number=self.mfa_serial_number,
                mfa_token=token_value if self.mfa_serial_number else None,
            )

            return role.credentials
        else:
            return self.sts.get_session_token(
                duration=self.token_duration,
                force_new=True,
                mfa_serial_number=self.mfa_serial_number,
                mfa_token=token_value if self.mfa_serial_number else None,
            )

    def clear_session(self):
        self.session = None
        self.session_expiration = datetime.min

    def get_session(self, token_value=None):
        if self.session_expired or token_value:
            try:
                while self.mfa_serial_number and not token_value:
                    token_value = self._prompt_token()

                self.session = self._create_session(token_value)

                # Needs testing, but might be worth subtracting X seconds
                # to prevent returning outdated session token. SDKs and AWS
                # may already handle this, so it may not be worth it.
                self.session_expiration = datetime.strptime(
                    self.session.expiration, "%Y-%m-%dT%H:%M:%SZ"
                )
            except Exception as ex:
                self.clear_session()
                raise ex

        return self.session



"""Console script for metamock."""

import sys

import click

from .config import write_config
from .server import Server
from .version import __timestamp__, __version__

header = f"{__name__.split('.')[0]} v{__version__} {__timestamp__}"


@click.command("metamock")
@click.version_option(message=header)
@click.option("-d", "--debug", is_flag=True, help="debug mode")
@click.option(
    "-c",
    "--config",
    type=click.Path(dir_okay=False),
    help="config file (default ~./metamock/config)",
)
@click.option("-h", "--host", type=str, default="169.254.169.254")
@click.option("-p", "--port", type=int, default=45000)
@click.option(
    "-P", "--profile", type=str, help="name of the profile to load by default"
)
@click.option(
    "-i", "--id", "key_id", type=str, help="configure: AWS_ACCESS_KEY_ID"
)
@click.option("-k", "--key", type=str, help="configure: AWS_SECRET_ACCESS_KEY")
@click.option("-r", "--region", type=str, help="configure: AWS_REGION")
@click.argument("command", type=str, default="server")
def cli(debug, config, host, port, profile, key_id, key, region, command):
    """serve mocked aws instance metadata

    Commands:

        configure: generate a config file at ~/.metamock/config

        server: run the server

    """

    def exception_handler(
        exception_type, exception, traceback, debug_hook=sys.excepthook
    ):

        if debug:
            debug_hook(exception_type, exception, traceback)
        else:
            click.echo(f"{exception_type.__name__}: {exception}", err=True)

    sys.excepthook = exception_handler

    click.echo(
        f"metamock {__version__} - because dev systems need 169.254.169.254 too."
    )

    if command == "configure":
        return write_config(
            key_id or "AWS_SECRET_KEY_ID",
            key or "AWS_SECRET_ACCESS_KEY",
            region or "AWS_REGION",
        )
    else:
        return Server(config, host, port, profile).run()


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



#!/usr/bin/env python

import re
import sys
from pathlib import Path

from .metadata import Metadata, Profile, bottle

APP_DIR = Path(__file__).parent
bottle.TEMPLATE_PATH.append(str(Path(APP_DIR) / "metadata" / "views"))
DEFAULT_CONFIG_FILE = Path(APP_DIR) / "metamock.config"
sys.path.append(APP_DIR)


class Server:
    def __init__(self, config, host, port, profile):

        self.host = host
        self.port = port
        self.profile = profile

        self.app = bottle.default_app()

        self.app.config.load_config(config or DEFAULT_CONFIG_FILE)

        user_config = Path.home() / ".metamock"
        if user_config.is_file():
            self.app.config.load_config(str(user_config))

        profile_name = profile or self.app.config.get(
            "metadata.profile", "default"
        )

        self.app.config.meta_set(
            "metadata",
            "obj",
            Metadata(self.parse_profiles(self.app.config), profile_name),
        )

    def run(self):
        return self.app.run(
            host=self.host
            or self.app.config.get("metadata.host", "169.254.169.254"),
            port=self.port or int(self.app.config.get("metadata.port", 45000)),
        )

    #    def port(self, value):
    #        value = int(value)
    #
    #        if value < 0 or value > 65535:
    #            raise argparse.ArgumentTypeError(
    #                "invalid port value: {}: value must be 0-65535".format(value))
    #
    #        return value
    #
    #
    #    def existing_file(self, value):
    #        if not Path(value).isfile():
    #            raise argparse.ArgumentTypeError(
    #                'file does not exist: {}'.format(value))
    #
    #        return value
    #
    #
    def to_int(self, value):
        return value if value is None else int(value)

    def to_bool(self, value, default=False):
        if value is None:
            return default

        return value.lower() in ("yes", "true", "y", "t", "1")

    def parse_profiles(self, config):
        prog = re.compile("^profile:([^\\.]+)\\.(.*)$")
        profiles = {}
        region = self.app.config.get("aws.region", "us-east-1")
        access_key = self.app.config.get("aws.access_key")
        secret_key = self.app.config.get("aws.secret_key")
        mfa_enabled = self.app.config.get("aws.mfa_enabled")
        mfa_secret = self.app.config.get("aws.mfa_secret")
        token_duration = self.app.config.get("metadata.token_duration")
        role_arn = self.app.config.get("metadata.role_arn")

        for key, value in config.items():
            res = prog.match(key)

            if res:
                profiles.setdefault(res.group(1), {})[res.group(2)] = value

        profiles.setdefault("default", {})

        for profile in profiles.values():
            profile.setdefault("region", region)
            profile.setdefault("access_key", access_key)
            profile.setdefault("secret_key", secret_key)
            profile["token_duration"] = self.to_int(
                profile.setdefault("token_duration", token_duration)
            )
            profile.setdefault("role_arn", role_arn)
            profile.setdefault("mfa_secret", mfa_secret)
            profile["mfa_enabled"] = self.to_bool(
                profile.setdefault("mfa_enabled", mfa_enabled), default=True
            )

        result = {}
        for name, values in profiles.items():
            try:
                result[name] = Profile(**values)
            except Exception as ex:
                raise Exception("Error loading profile {}".format(name), ex)

        return result



#!/usr/bin/env python

"""Tests for `metamock` package."""

from click.testing import CliRunner

import metamock
from metamock import __version__, cli


def test_version():
    """Test reading version and module name"""
    assert metamock.__name__ == "metamock"
    assert __version__
    assert isinstance(__version__, str)


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli)
    assert result.exit_code != 0, result
    assert "metamock" in result.output

    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0, result
    assert "Show this message and exit." in result.output



#!/usr/bin/python

import os

from Tkinter import ACTIVE, Button, Entry, Label, Tk

root = Tk()
root.wm_title("Enter MFA Token")

Label(root, text="Token").pack()
entry = Entry(root)
entry.pack(padx=5)


def done():
    print(entry.get())
    root.destroy()


b = Button(root, text="OK", default=ACTIVE, command=done)
b.pack(pady=5)

entry.focus_force()
root.bind("<Return>", (lambda e, b=b: b.invoke()))
os.system(
    (
        """/usr/bin/osascript -e 'tell app "Finder" to set """
        """frontmost of process "Python" to true' """
    )
)
root.mainloop()



#!/usr/bin/env python
import os
import os.path

OVERWRITE = True


def write_config(access_key=None, secret_key=None, region=None):
    metadata_dir = os.path.join(os.path.expanduser("~"), ".metamock")

    if not os.path.isdir(metadata_dir):
        os.mkdir(metadata_dir)

    key_map = {
        "AWS_ACCESS_KEY_ID": "access_key",
        "AWS_SECRET_ACCESS_KEY": "secret_key",
        "AWS_REGION": "region",
    }
    config_file = os.path.join(metadata_dir, "config")
    if OVERWRITE or not os.path.isfile(config_file):
        ofp = open(config_file, "w")
        ofp.write("[aws]\n")
        for k, v in os.environ.items():
            if k in key_map:
                ofp.write(key_map[k] + "=" + v + "\n")
        ofp.write("mfa_enabled=False\n")
        ofp.write("[metadata]\n")
        ofp.write("profile=default\n")
        ofp.close



from boto.exception import BotoServerError

from .bottle import delete, post, request, response, route, view


@route("/latest/meta-data/local-hostname")
def localhost():
    response.content_type = "text/plain; charset=UTF-8"
    return "localhost"


@route("/latest/meta-data/hostname")
def host():
    response.content_type = "text/plain; charset=UTF-8"
    return "localhost"


@route("/latest/meta-data/public-hostname")
def public_host():
    response.content_type = "text/plain; charset=UTF-8"
    return "localhost"


@route("/latest/meta-data/mac")
def mac_addr():
    response.content_type = "text/plain; charset=UTF-8"
    return "mac1234"


@route("/latest/meta-data/network/interfaces/macs/mac1234/vpc-id")
def vpc_id():
    response.content_type = "text/plain; charset=UTF-8"
    return "vpc-1234"


@route("/latest/meta-data/local-ipv4")
def localip4():
    response.content_type = "text/plain; charset=UTF-8"
    return "127.0.1.1"


@route("/latest/meta-data/instance-type")
def instance_type():
    response.content_type = "text/plain; charset=UTF-8"
    return "r3.2xlarge"


@route("/latest/meta-data/iam/security-credentials")
@route("/latest/meta-data/iam/security-credentials/")
def list_profiles():
    response.content_type = "text/plain; charset=UTF-8"
    return "local-credentials"


@route("/latest/meta-data/iam/security-credentials/local-credentials")
def get_credentials():
    try:
        session = request.app.config.meta_get("metadata", "obj").get_session()

        return {
            "Code": "Success",
            "AccessKeyId": session.access_key,
            "SecretAccessKey": session.secret_key,
            "Token": session.session_token,
            "Expiration": session.expiration,
        }
    except BotoServerError as e:
        response.status = e.status
        return {"error": {"message": e.message}}


@route("/latest/dynamic/instance-identity/document")
def get_identity_document():
    return {
        "privateIp": "127.0.0.1",
        "devpayProductCodes": None,
        "availabilityZone": "us-east-1a",
        "version": "2010-08-31",
        "instanceId": "i-12345678",
        "billingProducts": None,
        "instanceType": "t2.nano",
        "accountId": "123456789012",
        "architecture": "x86_64",
        "kernelId": None,
        "ramdiskId": None,
        "imageId": "ami-12345678",
        "pendingTime": "2016-01-01T00:00:00Z",
        "region": "us-east-1",
    }


@route("/manage")
@view("manage")
def manage():
    metadata = request.app.config.meta_get("metadata", "obj")

    return {
        "session": metadata.session,
        "profile_name": metadata.profile_name,
        "profiles": metadata.profiles,
    }


@route("/manage/profiles")
def get_profiles():
    metadata = request.app.config.meta_get("metadata", "obj")

    return {
        "profiles": [
            _profile_info(metadata, name, profile)
            for name, profile in metadata.profiles.items()
        ]
    }


@route("/manage/profiles/<name>")
def get_profile(name):
    metadata = request.app.config.meta_get("metadata", "obj")

    if name not in metadata.profiles:
        response.status = 404
        return {"error": {"message": "profile does not exist"}}

    return {"profile": _profile_info(metadata, name, metadata.profiles[name])}


@route("/manage/session")
def get_session():
    metadata = request.app.config.meta_get("metadata", "obj")
    result = {
        "profile": _profile_response(metadata.profile_name, metadata.profile)
    }

    if not metadata.session_expired:
        result["session"] = _session_response(metadata.session)

    return result


@delete("/manage/session")
def delete_session():
    request.app.config.meta_get("metadata", "obj").clear_session()


@post("/manage/session")
def create_session():
    metadata = request.app.config.meta_get("metadata", "obj")
    token = _get_value(request, "token")
    profile = _get_value(request, "profile")

    if not token and not profile:
        response.status = 400
        return {"error": {"message": "token and/or profile is required"}}

    if profile:
        metadata.profile_name = profile

    if token:
        try:
            request.app.config.meta_get("metadata", "obj").get_session(token)
        except BotoServerError as e:
            response.status = e.status
            return {"error": {"message": e.message}}

    return get_session()


def _get_value(request, key):
    value = request.forms.get(key)

    if value is None and request.json:
        value = request.json.get(key)

    return value


def _session_response(session):
    return {
        "accessKey": session.access_key,
        "secretKey": session.secret_key,
        "sessionToken": session.session_token,
        "expiration": session.expiration,
    }


def _profile_info(metadata, name, profile):
    result = _profile_response(name, profile)

    if not profile.session_expired:
        result["session"] = _session_response(profile.session)

    if metadata.profile_name == name:
        result["active"] = True

    return result


def _profile_response(name, profile):
    response = {"accessKey": profile.access_key, "region": profile.region}

    if profile.token_duration:
        response["tokenDuration"] = profile.token_duration

    if profile.role_arn:
        response["roleArn"] = profile.role_arn

    return response



import base64
import hashlib
import hmac
import time
from datetime import datetime


class Totp(object):
    def __init__(
        self, secret, interval_secs=30, digits=6, digest=hashlib.sha1
    ):
        """
        Create a new TOTP code generator.

        Parameters:
         secret (string|list of byte): shared secret as either a base32 encoded string or byte array
         interval_secs (int): interval, in seconds, to generate codes at
         digits (int): number of digits in the generated codes
         digest (function): HMAC digest function to use
        """
        if isinstance(secret, str):
            secret = secret_to_bytes(secret)

        self.secret = secret
        self.interval_secs = interval_secs
        self.digits = digits
        self.digest = digest

    def generate(self, at=None):
        """
        Generate a new OTP code.

        Parameters:
         at (datetime): timestamp to generate the code for or None to use current time

        Returns:
         (int): generated code
        """
        timecode = self.__timecode(at or datetime.now())
        hmac_hash = hmac.new(self.secret, timecode, self.digest).digest()

        offset = ord(hmac_hash[19]) & 0xF
        code = (
            (ord(hmac_hash[offset]) & 0x7F) << 24
            | (ord(hmac_hash[offset + 1]) & 0xFF) << 16
            | (ord(hmac_hash[offset + 2]) & 0xFF) << 8
            | (ord(hmac_hash[offset + 3]) & 0xFF)
        )

        return code % 10 ** self.digits

    def __timecode(self, at):
        return timestamp_to_bytestring(
            int(time.mktime(at.timetuple()) / self.interval_secs)
        )


def secret_to_bytes(secret):
    """Convert base32 encoded secret string to bytes"""
    return base64.b32decode(secret)


def timestamp_to_bytestring(val, padding=8):
    """Convert Unix timestamp to bytes"""
    result = []

    while val != 0:
        result.append(chr(val & 0xFF))
        val = val >> 8

    return "".join(reversed(result)).rjust(padding, "\0")



from __future__ import absolute_import

from functools import reduce


def get_value(d, *keys):
    return reduce(lambda a, b: a[b], keys, d)


def first_item(items):
    """Returns first item in the list or None if empty."""
    return items[0] if len(items) > 0 else None


class cache(object):
    """Computes attribute value and caches it in the instance.
    Python Cookbook (Denis Otkidach)
        http://stackoverflow.com/users/168352/denis-otkidach
    This decorator allows you to create a property which can be computed once
    and accessed many times. Sort of like memoization.
    """

    def __init__(self, method, name=None):
        # record the unbound-method and the name
        self.method = method
        self.name = name or method.__name__
        self.__doc__ = method.__doc__

    def __get__(self, inst, cls):
        if inst is None:
            # instance attribute accessed on class, return self
            # You get here if you write `Foo.bar`
            return self
        # compute, cache and return the instance's attribute value
        result = self.method(inst)
        # setattr redefines the instance's attribute so this doesn't get called
        # again
        setattr(inst, self.name, result)
        return result
